/* xml.c */
/* Norbert Doerry 

   Version 1.0
   March 2000

   version 1.0a
   May 2008
   fixed minor bug that strip last character of the value being read in 
   fixed debug message that stated hitting an end tag was unanticipate -- it was


   version 1.0b
   May 2008

   added routine 


    XMLENTITY *extract_XMLENTITY_name(XMLENTITY *x,char *s)

   version 1.1
   May 2017

   added routine

   void name2lower_XMLENTITY(XMLENTITY *x,int recursive_flag)

    Copyright (C) 2008 Norbert H. Doerry

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


    if you find any bugs in this software, please email doerry@aol.com

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "xml.h"

static int debug = 0;

/* local routines */

char *xmlcopystring(char *);
void xmlStringStrip(char *);
XMLENTITY *xmlParseTag(char *);
XMLENTITY *fget_XMLENTITY(FILE *,int *);


void XMLsetDEBUG(void)
{
 extern int debug;
 
 debug = 1;
}

void XMLclearDEBUG(void)
{
 extern int debug;
 
 debug = 0;
}

/* alloc_XMLENTITY(void) allocates an XMLENTITY and initializes the default
   values,  the name and object contents are left as void */

XMLENTITY *alloc_XMLENTITY(void)
{
 XMLENTITY *xml;
 
 xml = (XMLENTITY *) calloc((size_t)1, sizeof(XMLENTITY));

 if (xml == (XMLENTITY *) NULL)
  {
   fprintf(stderr," *** ERROR: Out of memory in alloc_XMLENTITY()\n");
   exit(EXIT_FAILURE);
  }

 xml->name = (char *) NULL;
 xml->s    = (char *) NULL;
 xml->type = ENTITY_UNKNOWN;


 xml->child = (XMLENTITY *) NULL;
 xml->Nmtoken = (XMLENTITY *) NULL;

 xml->next = (XMLENTITY *) NULL;
}



void free_XMLENTITY(XMLENTITY *xml)
{

  if (xml == (XMLENTITY *) NULL) return;
   
  if (xml->name != (char *) NULL) free((void *) xml->name);
  
  if (xml->s != (char *) NULL) free((void *) xml->s);

  if (xml->child != (XMLENTITY *) NULL)
   free_XMLENTITY(xml->child);

  if (xml->Nmtoken != (XMLENTITY *) NULL)
   free_XMLENTITY(xml->Nmtoken);

  if (xml->next != (XMLENTITY *) NULL)
   free_XMLENTITY(xml->next);

  free((void *) xml);
}


XMLENTITY *readfile_XMLENTITY(FILE *in,int *errflag)
{

 char ch;
 XMLENTITY *xml,*x1,*x2;
 int err;
 extern int debug;
 int i;

 xml = (XMLENTITY *) NULL;

 if (errflag == (int *) NULL || in == (FILE *) NULL)
  {
   fprintf(stderr," *** ERROR: NULL Pointers on entry to readfile_XMLENTITY\n");
   return xml;
  }

 if (debug) printf("[Entering readfile_XMLENTITY: *errflag = %d\n",*errflag);

 *errflag = ERROR_NONE;  /* initialize to zero */

 while (1)
  {
 
   /* look for the start of the tag */

   do
    {
     ch = (char) fgetc(in);
    }
   while (ch != '<' && ch != EOF); 

   if (ch == EOF) 
    {
     if (debug) printf("[Exiting readfile_XMLENTITY: *errflag = %d]\n",*errflag);
     return xml;  /* EOF Reached before another entity found - done */  
    }

   /* read the entity */

   err = ERROR_NONE;

   x1 = fget_XMLENTITY(in,&err);

   if (debug) printf("[return from fget_XMLENTITY  err = %d]\n", err);

   /* attach to linked list */

   if (x1 != (XMLENTITY *) NULL)
    {
     if (xml == (XMLENTITY *) NULL)
      xml = x1;
     else
      x2->next = x1;
 
     x2 = x1;

     /* store the error flag */

     x1->errflag = err;

     if (err != ERROR_NONE) *errflag = err;

     if (debug) printf("[*errflag = %d]\n",*errflag);
    }

  } 
  
}

#define DEFAULT_MAXLEN 2048


/* following routine assumes that the '<' for the tag has already been read in */


XMLENTITY *fget_XMLENTITY(FILE *in,int *errflag)
{
 char *buf,ch;
 XMLENTITY *xml,*x1,*x2;
 long int maxlen;
 int exit_state;
 int i,j;
 int flag;
 int err;
 extern int debug;

 xml = (XMLENTITY *) NULL;

 if (errflag == (int *) NULL)
  {

   fprintf(stderr," *** ERROR: NULL errflag pointer in fget_XMLENTITY()\n");
   exit(EXIT_FAILURE);
  }
 
 if (in == (FILE *) NULL)
  {
   fprintf(stderr," *** ERROR: NULL File Pointer in fget_XMLENTITY()\n");
   exit(EXIT_FAILURE);
  }

 if (debug) printf("{Entering fget_XMLENTITY: *errflag = %d}\n",*errflag);

 /* allocate the buffer */

 maxlen = DEFAULT_MAXLEN;

 buf = (char *) calloc((size_t)maxlen,sizeof(char));
 if (buf == (char *) NULL)
  { 
   fprintf(stderr," *** ERROR: Out of memory in fget_XMLENTITY() 1\n");
   exit(EXIT_FAILURE);
  }


 /* read in the next character */

 buf[0] = (char) fgetc(in);

 /* see if an end tag */

 if (buf[0] == '/')
  {
   *errflag = ERROR_ENDTAG;
   free((void *) buf);

   if (debug) 
    printf("{Exiting fget_XMLENTITY with endtag: *errflag = %d}\n",*errflag);

   return xml;
  }


 /* see if a possible comment or CDATA section or other stuff*/
 
 if (buf[0] == '!')
  {
   /* get the next character */

   buf[0] = (char) fgetc(in);

   /* see if the element is empty */

   if (buf[0] == '>')
    {
     *errflag = ERROR_SYNTAX;
     free((void *) buf);

     if (debug) 
      printf("{Exiting fget_XMLENTITY empty tag: *errflag = %d}\n",*errflag);

     return xml;  /* return NULL Pointer */
    }


   /* see if a possible comment */

   if (buf[0] == '-')
    {

     /* get the next character */

     buf[0] = (char) fgetc(in);
  
     /* confirm that its a comment */


     if (buf[0] == '-')
      {
       /* read in comment until "-->" is reached */
         
       if (debug) 
        printf("{fget_XMLENTITY reading comment: *errflag = %d}\n",*errflag);

       for (i = 0,exit_state = 0 ; i < maxlen && exit_state != 3 ; i++)
	{
         buf[i] =  fgetc(in);

         if (buf[i] == EOF)
	  {
	   /* error */

           free((void *) buf);
	   *errflag = ERROR_EOF;

          if (debug) 
           printf("{Exiting fget_XMLENTITY unanticipated EOF: *errflag = %d}\n",*errflag);
 
           return xml; /* return NULL Pointer */
          }
          
         /* see if done */

         if (exit_state == 0 && buf[i] == '-') exit_state = 1;
	 else if (exit_state == 1 && buf[i] == '-') exit_state = 2;
	 else if (exit_state == 2 && buf[i] == '>') exit_state = 3;
	 else exit_state = 0;
        }

       /* terminate the comment */
       buf[i-3] = (char) NULL;

       if (i == maxlen) /* see if filled the buffer */
        {
         *errflag = ERROR_BUFFER_OVERRUN;
	 free((void *) buf);

         if (debug) 
          printf("{fget_XMLENTITY in comment, buffer overrun: *errflag = %d}\n",*errflag);

         return xml;  /* return NULL Pointer */
        }

       /* return an xml structure with the data in the buffer */	

       xml = alloc_XMLENTITY();

       xml->name = xmlcopystring("COMMENT");
       xml->s    = xmlcopystring(buf);
       xml->type = ENTITY_COMMENT;

       if (debug) 
        printf("{fget_XMLENTITY in comment, comment done: *errflag = %d}\n{%s}\n",
			*errflag,xml->s);
         
       free((void *) buf);
       return xml;

      }          
     else  /* its not a real comment */
      {

       /* set the error flag */

       *errflag = ERROR_SYNTAX;
 
       free((void *) buf);

       if (debug) 
        printf("{fget_XMLENTITY in comment, not quite a comment: *errflag = %d}\n",
		*errflag);

       return xml; /* return NULL pointer */

      }

    }  /* end of comment section */


   /* see if a possible CDATA Section  <![CDATA[  (if so, then ignore the section) */

   else if (buf[0] == '[')  
    {
     /* check for  <![CDATA[  */

     for (i = 1; i < 7 ; i++)
      {
       buf[i] = (char) fgetc(in);
       if (buf[i] == EOF)
	{
	 /* error */

         *errflag = ERROR_EOF;
	 free((void *) buf);
         if (debug) 
          printf("{Exiting fget_XMLENTITY unanticipated EOF: *errflag = %d}\n",*errflag);
 
	 return xml; /* return NULL Pointer */
        }
      } 

     buf[7] = (char) NULL;

     if (strcmp(buf,"[CDATA[") != 0)
      {
       /* error */

       *errflag = ERROR_SYNTAX;

       if (debug) 
        printf("{Exiting fget_XMLENTITY not CDATA (%s): *errflag = %d}\n",buf,*errflag);
 
       free((void *) buf);
       return xml; /* return NULL Pointer */
      }

     /* at this point, we have the beginning of a CDATA section
        need to read in until have the ']]>' delimiter or until 
        the buffer length is exceeded */

     if (debug) 
      printf("{Reading CDATA in fget_XMLENTITY: *errflag = %d}\n",*errflag);
  
          
     for (i = 0,exit_state = 0 ; i < maxlen && exit_state != 3 ; i++)
      {
       buf[i] =  fgetc(in);

       if (buf[i] == EOF)
        {
	 /* error */

         free((void *) buf);
         *errflag = ERROR_EOF;
         return xml; /* return NULL Pointer */

        }

       /* see if done */

       if (exit_state == 0 && buf[i] == ']') exit_state = 1;
       else if (exit_state == 1 && buf[i] == ']') exit_state = 2;
       else if (exit_state == 2 && buf[i] == '>') exit_state = 3;
       else exit_state = 0;
      }

     buf[i-3] = (char) NULL; /* terminate the CDATA */

     if (i == maxlen) /* see if filled the buffer */
      {
       *errflag = ERROR_BUFFER_OVERRUN;
       free((void *) buf);
 
       if (debug) 
        printf("{Buffer overrun Reading CDATA in fget_XMLENTITY: *errflag = %d}\n",
		*errflag);

       return xml;  /* return NULL Pointer */
      }

     /* return an xml structure with the data in the buffer */	

     xml = alloc_XMLENTITY();

     xml->name = xmlcopystring("CDATA");
     xml->s    = xmlcopystring(buf);
     xml->type = ENTITY_CDATA;
    
     free((void *) buf);

     if (debug) 
      printf("{Done Reading CDATA in fget_XMLENTITY: *errflag = %d}\n",*errflag);

     return xml;
    }  /* end of CDATA section */

   /* otherwise its some other command -- read until end of tags*/

   for (i = 0, exit_state = 1 ; exit_state > 0 && i < maxlen ; i++)
    {

      buf[i] = (char) fgetc(in);

      if (buf[i] == EOF)
       {
	/* ERROR */

	 *errflag = ERROR_EOF;

	 free((void *) buf);

         if (debug) 
          printf("{EOF Reading other command in fget_XMLENTITY: *errflag = %d}\n",
		  *errflag);


	 return xml; /* return NULL pointer */

       }

      /* update the exit_state which corresponds to the number
	 of '>' still needed to match all the '<' */

      else if (buf[i] == '<')
       exit_state++;
      else if (buf[i] == '>')
       exit_state--;

     }

    if (i == maxlen) /* see if filled the buffer */
     {
      *errflag = ERROR_BUFFER_OVERRUN;
      free((void *) buf);

      if (debug) 
       printf("{Buffer overrun Reading other cmd in fget_XMLENTITY: *errflag = %d}\n",
		*errflag);

      return xml;  /* return NULL Pointer */
     }

 
    buf[i-1] = (char) NULL;

    /* return an xml structure with the data in the buffer */	

   xml = alloc_XMLENTITY();

   xml->name = xmlcopystring("UNKNOWN");
   xml->s    = xmlcopystring(buf);
   xml->type = ENTITY_UNKNOWN;

   *errflag = ERROR_SYNTAX;
   free((void *) buf);

   if (debug) 
    printf("{Done reading other command: *errflag = %d}\n",
		*errflag);

   return xml;
  }


 /* see if a possible Processing Instruction (PI) <? info ?> */

 else if (buf[0] == '?')
  {
   for (i = 0, exit_state = 0 ; exit_state != 2 && i < maxlen ; i++)
    {

      buf[i] = (char) fgetc(in);

      if (buf[i] == EOF)
       {
	/* ERROR */

	 *errflag = ERROR_EOF;

	 free((void *) buf);

         if (debug) 
          printf("{EOF Reading PI in fget_XMLENTITY: *errflag = %d}\n",
		*errflag);

	 return xml; /* return NULL pointer */

       }

      /* update the exit_state which corresponds to the number
	 of '>' still needed to match all the '<' */

      else if (buf[i] == '?' && exit_state == 0) exit_state = 1;
      else if (buf[i] == '>' && exit_state == 1) exit_state = 2;

     }

    if (i == maxlen) /* see if filled the buffer */
     {
      *errflag = ERROR_BUFFER_OVERRUN;
      free((void *) buf);

      if (debug) 
       printf("{Buffer overrun Reading PI in fget_XMLENTITY: *errflag = %d}\n",
		*errflag);

      return xml;  /* return NULL Pointer */
     }

 
    buf[i-2] = (char) NULL;

    /* return an xml structure with the data in the buffer */	

   xml = alloc_XMLENTITY();

   xml->name = xmlcopystring("PROCESSING INSTRUCTION");
   xmlStringStrip(buf);
   xml->s    = xmlcopystring(buf);
   xml->type = ENTITY_PI;

   free((void *) buf);

   if (debug) 
    printf("{Done Reading PI in fget_XMLENTITY: *errflag = %d}\n",
		*errflag);


   return xml;
    
  }


 /* read in the tag name */


 i = 0;
  
 do
  {
   i++;
   buf[i] = fgetc(in);

   if (buf[i] == EOF)
    { 
     *errflag = ERROR_EOF;
     free((void *) buf);
     return xml;  /* return NULL pointer */
    }
  }
 while (buf[i]!= '>' && i < maxlen-1);

 buf[i] = (char) NULL; /* NULL TERMINATE */

 if (i == maxlen - 1) /* see if filled the buffer */
  {
   *errflag = ERROR_BUFFER_OVERRUN;
   free((void *) buf);

   if (debug) 
    printf("{Buffer overrun Reading tag in fget_XMLENTITY: *errflag = %d}\n",
		*errflag);

   return xml;  /* return NULL Pointer */
  }


 /* see if the tag is an empty element tag */ 


 if (buf[i-1] == '/')
  {
    buf[i-1] = (char) NULL;
    xml = xmlParseTag(buf);
    xml->type = ENTITY_DATA;
    free((void *) buf);

    if (debug) 
     printf("{Empty Element tag <%s> in fget_XMLENTITY: *errflag = %d}\n",
		xml->name, *errflag);


    return xml;
  }

  /* it is a start tag */

 else 
  {
    xml = xmlParseTag(buf);  /* parse the tag */
    xml->type = ENTITY_DATA;

    if (debug) 
     printf("{Start tag <%s> in fget_XMLENTITY: *errflag = %d}\n",
		xml->name, *errflag);


    x2 = (XMLENTITY *) NULL; /* pointer to last child entity */

    /* read in the entity */

    for (i = 0 ; i < maxlen ; i++)
     {
      buf[i] = fgetc(in);

      if (buf[i] == EOF)
       { 
        *errflag = ERROR_EOF;
        buf[i] = (char) NULL;
        xml->s = xmlcopystring(buf);
        free((void *) buf);

        if (debug) 
         printf("{EOF in tag <%s> in fget_XMLENTITY: *errflag = %d}\n",
		xml->name, *errflag);


        return xml;  /* return xml structure the way it is */
       }
      
      /* see if it's a tag */

      else if (buf[i] == '<')
       {

        /* i--; */ /* -- should not -- ignore the '<' */

        err = 0;

        x1 =  fget_XMLENTITY(in,&err);  /* recursion here */
          
        if (x1 != (XMLENTITY *) NULL)
         {        

          if (xml->child == (XMLENTITY *) NULL)
           xml->child = x1;
          else
           x2->next = x1;

          x2 = x1;
          if (err != ERROR_NONE) *errflag = err;

          if (debug) 
           printf("{child tag <%s> to <%s> in fget_XMLENTITY: *errflag = %d}\n",
  		x1->name,xml->name, *errflag);

         }
        else if (err == ERROR_ENDTAG)
         {
           /* tag is an end tag */           

          for (j=0 ; j < strlen(xml->name) ; j++)
            {
             ch = fgetc(in);
             if (ch != xml->name[j])
              {
               *errflag = (ch == EOF) ? ERROR_EOF : ERROR_SYNTAX;
               buf[i] = (char) NULL;
               xml->s = xmlcopystring(buf);
               free((void *) buf);

               if (debug) 
                printf("{wrong endtag (j=%d)to <%s> in fget_XMLENTITY: *errflag = %d}\n",
  	   		j,xml->name, *errflag);


               return xml;
              }
             
            }
     
           /* if we are this far, the end tag is correct up to the name of the
              tag.  remaining characters should be spaces terminated with a > */

           do
            {
              ch = fgetc(in);
            }

           while (isspace(ch) != 0);

           /* ch should now be the > */

           if (ch != '>')
            {
             *errflag = (ch == EOF) ? ERROR_EOF : ERROR_SYNTAX;
             if (debug) 
              printf("{improper end of tag <%s> in fget_XMLENTITY: *errflag = %d}\n",
  		xml->name, *errflag);
            }
 
          
           buf[i] = (char) NULL;
	   xmlStringStrip(buf);
           xml->s = xmlcopystring(buf);
           free((void *) buf);

          if (debug) 
           printf("{End of tag <%s> in fget_XMLENTITY: *errflag = %d}\n{%s}\n",
  		xml->name, *errflag,xml->s);


           return xml;
         } 

       }  /* end of it's a tag */ 
      
     }  /* end of reading in the entity */ 

  } /* end of its a start tag */
   
}

/* creates an xml entity structure and populates the Name Tokens */

XMLENTITY *xmlParseTag(char *s)
 {
  XMLENTITY *x1, *x2;
  char ch;
  int i1,i2,i3;

  /* error checking */

  if (s == (char *) NULL)
   return (XMLENTITY *) NULL;

  /* allocate the XML structure */

  x1 = alloc_XMLENTITY();
 
  /* find the start of the name */

  for (i1 = 0; isspace(s[i1]) != 0 ; i1++);

  /* find the end of the name */

  for (i2 = i1 ; isspace(s[i2]) == 0 && s[i2] != (char) NULL ; i2++);

  /* save s[i2] for NULL checking */

  ch = s[i2];

  /* terminate s at i2 */

  s[i2] = (char) NULL;

  x1->name = xmlcopystring(s+i1);  

  /* see if no name tokens and no trailing spaces */

  if (ch == (char) NULL)
   return x1; 
  
  /* find the Name Tokens */

 while (1)
  {
  
   for (i1 = i2+1 ; isspace(s[i1]) != 0 ; i1++);

   if (s[i1] == (char) NULL)  /* see if no more name tokens */
    break;

   /* find the end of the name of the token */

   for (i2 = i1 ; s[i2] != '=' && s[i2] != (char) NULL ; i2++);

   if (x1->Nmtoken == (XMLENTITY *) NULL)
    {
     x1->Nmtoken = alloc_XMLENTITY();
     x2 = x1->Nmtoken;
    }
   else
    {
     x2->next = alloc_XMLENTITY();
     x2 = x2->next;
    }

   ch = s[i2];  /* save the character, its either an '=' or NULL */

   s[i2] = (char) NULL;  /* make it a NULL in order to copy it */

   /* work backwards to get rid of the spaces */

   for (i3 = i2-1 ; isspace(s[i3]) != 0 && i3 > i1 ; i3--)
    s[i3] = (char) NULL;

   x2->name = xmlcopystring(s+i1);

   if (ch == (char) NULL)  /* if it was a NULL, then we're done */
    break;
    
   /* if it was an equal sign, then get the value */

   for (i1 = i2+1 ; isspace(s[i1]) != 0 ; i1++);

   if (s[i1] == (char) NULL)  /* no actual value provided */
    break;
   
   ch = s[i1];
   if (ch != '"') ch = ' ';  /* ch is the the delimiter */

   for (i2 = i1+1 ; s[i2] != (char) NULL ; i2++)
    {
     if (ch == '"' && s[i2] == ch)
      break;
     if (ch == ' ' && isspace(s[i2]) != 0)
      break;
    }

   ch = s[i2];  /* save this to test for NULL */

   s[i2] = (char) NULL;
   if (s[i1] == '"') i1++;

   x2->s = xmlcopystring(s+i1);  /* save the value */

   if (ch == (char) NULL)  /* see if we are done */
    break;
  }

 return x1;
}

char *xmlcopystring(char *s)
 {
   char *ss;

   /* error checking */

   if (s == (char *) NULL)
    return s;
 
   /* allocate the new array */

   ss = (char *) calloc((size_t) strlen(s)+1,(size_t)sizeof(char));
   if (ss == (char *) NULL)
    {
     fprintf(stderr," *** ERROR: Out of Memory in xmlcopystring()\n");
     exit(EXIT_FAILURE);
    }
  
   /* copy the string */

   return strcpy(ss,s);   
  
  }


void xmlfprint(FILE *out, XMLENTITY *xml, int indent)
 {

  int i; 
  XMLENTITY *x;

  if (out == (FILE *) NULL)
   out = stdout;

  if (xml == (XMLENTITY *) NULL)
   return;

  if (indent < 0) indent = 0;

  if (xml->type == ENTITY_DATA)
   {
    for (i = 0; i< indent ; i++) fputc(' ',out);
 
    fprintf(out,"<%s",xml->name);
    for (x = xml->Nmtoken ; x != (XMLENTITY *) NULL ; x = x->next)
     {
      if (x->name != (char *) NULL)
       {       
        fprintf(out," %s",x->name);
        if (x->s != (char *) NULL)
         fprintf(out,"=\"%s\"",x->s);
       }
     }
    fprintf(out,">[ENTITY_DATA errflag = %d]\n",xml->errflag);
  
    if (xml->s != (char *) NULL)
     {
      for (i = 0; i< indent ; i++) fputc(' ',out); 
      fprintf(out,"{%s}\n",xml->s);
     }

    for (x = xml->child ; x != (XMLENTITY *) NULL ; x = x->next)
     xmlfprint(out,x,indent+1);

    for (i = 0; i< indent ; i++) fputc(' ',out);
    fprintf(out,"</%s>\n",xml->name);
    
   }

  else if (xml->type == ENTITY_PI)
   {
    for (i = 0; i< indent ; i++) fputc(' ',out);
    fprintf(out,"<?%s?>\n",xml->s);
   }

  else if (xml->type == ENTITY_CDATA)
   {
    for (i = 0; i< indent ; i++) fputc(' ',out);
    fprintf(out,"<![CDATA[%s]]>\n",xml->s);
 
   }
  else if (xml->type == ENTITY_COMMENT)
   {
    for (i = 0; i< indent ; i++) fputc(' ',out);
    fprintf(out,"<!-- %s -->\n", xml->s);
 
   }
   
  
 }   


void xmlStringStrip(char *s)
{
  char *in, *out;


  /* error checking */

  if (s == (char *) NULL)
   return;

  /* ignore leading white spaces  */

  for (in = s ; isspace(*in) != 0 ; in++);

  /* shift everything left */

  for (out = s ; *in != (char) NULL ; in++,out++)
   *out = *in;

  *out = (char) NULL;

  /* ignore trailing spaces */

  for (out-- ; isspace(*out) != 0 && out > s ; out--)
   *out = (char) NULL;

}       
  
/* looks for a child entity of x with name in s and returns the child entity or NULL if not found */

XMLENTITY *extract_XMLENTITY_name(XMLENTITY *x,char *s)
{
  
  XMLENTITY *px;

  if (x == (XMLENTITY *) NULL || s == (char *) NULL)
   return (XMLENTITY *) NULL;


  for (px = x->child; px != (XMLENTITY *) NULL ; px = px->next)
   {
    if (strcmp(px->name,s) == 0)
     return px;
   }

  return px;

}

void xmlstr2lower(char *);
 

/* converts the entity name to lower case.  If recursive_flag is != 0 also applied to all child entities   */

void name2lower_XMLENTITY(XMLENTITY *x,int recursive_flag)
{
 XMLENTITY *px;
 
 if (x == (XMLENTITY *) NULL)
  return;


 xmlstr2lower(x->name);
 

 if (recursive_flag != 0)
  {
   for (px = x->child ; px != (XMLENTITY *) NULL ; px = px->next)
    name2lower_XMLENTITY(px,recursive_flag);
  }


}

void xmlstr2lower(char *s)
 {
   char *ss;

   for (ss = s  ;*ss != '\0' ; ss++)
     *ss = (char)tolower((int) *ss); 

 }

 

  
  
